/*
 *  LinePol - ARTIST
 *
 *  Created by Rolf Kuiper
 *
 *    All rights reserved.
 *
 */

#include "lime.h"
#if LINEPOLARIZATION == ON
#include "linepolarization.h"



int *linepolarization_SplittedLevels;
int linepolarization_NSplittedLevels;



// Output: Boolean
int linepolarization_AreMagneticSublevelsTreatedExplicitly(int J){

	int flag = 0;
	int i;
	for(i = 0; i < linepolarization_NSplittedLevels; i++){
		if(linepolarization_SplittedLevels[i]   == J) flag = 1;
		if(linepolarization_SplittedLevels[i]+1 == J) flag = 1;
	}

	return flag;
}



// TODO: Improve the performance of this search algorithm
// Output: Energy level number (zero-indexed) corresponding to the given input J and M
int linepolarization_GetEnergyLevelNumber(int J, int M, molData *m, int ispec){

	int ilev;

	for(ilev = J; ilev < m[ispec].nlev; ilev++){
		if(J == m[ispec].J[ilev] && M == m[ispec].M[ilev]) break;
	}

	return ilev;
}



double linepolarization_GetClebschGordonCoefficient(int Jup, int Mup, int Mdown){

	int dM;
	dM = Mup - Mdown;
	
	// cast input integers to double:
	double J = (double) Jup;
	double M = (double) Mup;

	// compute and return the Clebsch-Gordon coefficient:
	switch(dM){
	case 1:
		return sqrt((J+M)*(J+M-1.0) / (2.0*J*(2.0*J-1.0)));
		break;
	case 0:
		return sqrt((J*J - M*M)     /     (J*(2.0*J-1.0)));
		break;
	case -1:
		return sqrt((J-M)*(J-M-1.0) / (2.0*J*(2.0*J-1.0)));
		break;
	default:
		printf("ERROR ERROR ERROR ERROR ERROR ERROR ERROR ERROR ERROR\n");
		printf("ERROR\n");
		printf("ERROR   In routine linepolarization_GetClebschGordonCoefficient():\n");
		printf("ERROR   dM = Mup - Mdown is not equal to -1, 0, or +1.\n");
		printf("ERROR\n");
		printf("ERROR ERROR ERROR ERROR ERROR ERROR ERROR ERROR ERROR\n");
		exit(1);
	}
}



void linepolarization_GetPropagationDirection_photon(double *n, struct grid *g, int here, int dir){
	
	double absn;
	
	n[0] = g[here].dir[dir].x[0];
	n[1] = g[here].dir[dir].x[1];
	n[2] = g[here].dir[dir].x[2];
	
	// normalize the propagation direction vector of the light:
	absn = sqrt(n[0]*n[0] + n[1]*n[1] + n[2]*n[2]);
	n[0] /= absn;
	n[1] /= absn;
	n[2] /= absn;
	
    // check
	if(fabs(sqrt(n[0]*n[0] + n[1]*n[1] + n[2]*n[2])-1.) > 1e-14){
		printf("ERROR: Detected in checks in photon.c\n");
		printf("ERROR: n is not a unit vector.\n");
		printf("ERROR: abs(n) = %e\n", absn);
		printf("ERROR: n[0] = %e\n", n[0]);
		printf("ERROR: n[1] = %e\n", n[1]);
		printf("ERROR: n[2] = %e\n", n[2]);
		bail_out("BUG DETECTED in photon.c");
	}
	
	return;
}



void linepolarization_GetPropagationDirection_raytrace(double *n, image *img, int im){
	
	double theta, phi;
	
	theta = img[im].theta;
	phi   = img[im].phi;
    
    //
    // In Lime, the angle phi is defined as the rotation angle around the line-of-sight
    //  (after applying the inclination angle theta as a rotation around the x-axis).
    // Unfortunately, this does not allow to ray-trace the model in any arbitrary direction, hence, should be changed in the future.
    // Here, we deal with it by assuming phi=0 and rotate the input model instead if necessary.
    //
    if(phi != 0){
        printf("ERROR DETECTED IN INPUT\n");
        printf(" When using LinePol, the angle phi has to be set to zero.\n");
        printf(" Here: phi = %e != 0.\n", phi);
        printf(" Please check your model.c input file.\n");
        exit(1);
    }
	
//    n[0] = cos(theta)*sin(phi);
//    n[1] = sin(theta);
//    n[2] = cos(theta)*cos(phi);
    
    n[0] = 0.0;
    n[1] = sin(theta);
    n[2] = cos(theta);
		
    // check
	double absn = sqrt(n[0]*n[0] + n[1]*n[1] + n[2]*n[2]);
	if(fabs(absn-1.) > 1e-14){
		printf("ERROR: n is not a unit vector.\n");
		printf("ERROR: abs(n) = %e\n", absn);
		printf("ERROR: n[0] = %e\n", n[0]);
		printf("ERROR: n[1] = %e\n", n[1]);
		printf("ERROR: n[2] = %e\n", n[2]);
		bail_out("BUG DETECTED in raytrace.c");
	}
	
	return;
}



double linepolarization_GetAngleBetweenPropagationDirectionAndMagneticField(struct grid *g, int here, double *n){
    
    double B[3], Bdotn, absB, angle_nB;
    
    magneticfield(g[here].x[0], g[here].x[1], g[here].x[2], B);
    
    Bdotn    =      B[0]*n[0] + B[1]*n[1] + B[2]*n[2];
    absB     = sqrt(B[0]*B[0] + B[1]*B[1] + B[2]*B[2]);
    angle_nB = acos(Bdotn/absB);
    
    // check
    if(angle_nB < 0 || angle_nB > PI){
        printf("ERROR: Detected in checks in linepolarization_GetAngleBetweenPropagationDirectionAndMagneticField() in linepolarization.c\n");
        printf("ERROR: angle_nB is not correct.\n");
        printf("ERROR: angle_nB = %e\n", angle_nB);
        exit(1);
    }
    
    return angle_nB;
}



void linepolarization_PolarizationFrame2CartesianFrame(struct grid *g, int here, double *n, double snu_para, double snu_perp, double *snu_pol){
    
    double B[3], absB;
    double e_perp[3], e_para[3];
    double Bcrossn[3], absBcrossn, Bdotn;
    double deviation = 1e-10;
    
    // get magnetic field:
    magneticfield(g[here].x[0], g[here].x[1], g[here].x[2], B);
    
    // normalize the magnetic field vector:
    absB = sqrt(B[0]*B[0] + B[1]*B[1] + B[2]*B[2]);
    B[0] /= absB;
    B[1] /= absB;
    B[2] /= absB;
    
    // compute polarization vectors:
    //  They are perpendicular and parallel to the projection of the magnetic field vector B[]
    //  in the plane perpendicular to the propagation direction of the light n[].
    Bdotn = B[0]*n[0] + B[1]*n[1] + B[2]*n[2];
    Bcrossn[0] = B[1]*n[2] - B[2]*n[1];
    Bcrossn[1] = B[2]*n[0] - B[0]*n[2];
    Bcrossn[2] = B[0]*n[1] - B[1]*n[0];
    absBcrossn = sqrt(Bcrossn[0]*Bcrossn[0] + Bcrossn[1]*Bcrossn[1] + Bcrossn[2]*Bcrossn[2]);
    e_perp[0] = Bcrossn[0] / absBcrossn;
    e_perp[1] = Bcrossn[1] / absBcrossn;
    e_perp[2] = Bcrossn[2] / absBcrossn;
    e_para[0] = (B[0] - n[0]*Bdotn) / absBcrossn;
    e_para[1] = (B[1] - n[1]*Bdotn) / absBcrossn;
    e_para[2] = (B[2] - n[2]*Bdotn) / absBcrossn;
    
    double e_perpdote_para;
    e_perpdote_para = e_perp[0]*e_para[0] + e_perp[1]*e_para[1] + e_perp[2]*e_para[2];
    if(fabs(e_perpdote_para) > deviation){
        printf("ERROR: Detected in checks in linepolarization.c\n");
        printf("ERROR: e_perp is not perpendicular to e_para.\n");
        printf("ERROR: e_perp dot e_para = %e\n", e_perpdote_para);
        exit(1);
    }
    
    e_perp[0] = fabs(e_perp[0]);
    e_perp[1] = fabs(e_perp[1]);
    e_perp[2] = fabs(e_perp[2]);
    e_para[0] = fabs(e_para[0]);
    e_para[1] = fabs(e_para[1]);
    e_para[2] = fabs(e_para[2]);
    
    double e_para_tot, e_perp_tot;
    e_para_tot = e_para[0] + e_para[1] + e_para[2];
    e_perp_tot = e_perp[0] + e_perp[1] + e_perp[2];
    
    snu_pol[0]  = e_para[0]/e_para_tot * snu_para  + e_perp[0]/e_perp_tot * snu_perp;
    snu_pol[1]  = e_para[1]/e_para_tot * snu_para  + e_perp[1]/e_perp_tot * snu_perp;
    snu_pol[2]  = e_para[2]/e_para_tot * snu_para  + e_perp[2]/e_perp_tot * snu_perp;
    
    // special handling, if the light travels exactly along the magnetic field lines:
    if(absBcrossn < 1e-14){
        if(fabs(snu_para - snu_perp)/snu_para > 1e-14){
            printf("ERROR: Detected in PolarizationFrame2CartesianFrame() in linepolarization.c\n");
            printf("ERROR: The light is moving along the magnetic field lines, but snu_para != snu_perp.\n");
            printf("ERROR: fabs(snu_para - snu_perp)/snu_para = %e\n", fabs(snu_para - snu_perp)/snu_para);
            exit(1);
        }
        snu_pol[0]  = (n[1]+n[2]) / (n[0]+n[1]+n[2]) * 0.5 * (snu_para + snu_perp);
        snu_pol[1]  = (n[2]+n[0]) / (n[0]+n[1]+n[2]) * 0.5 * (snu_para + snu_perp);
        snu_pol[2]  = (n[0]+n[1]) / (n[0]+n[1]+n[2]) * 0.5 * (snu_para + snu_perp);
    }
    
    // checks
    absB = sqrt(B[0]*B[0] + B[1]*B[1] + B[2]*B[2]);
    if(fabs(absB-1.) > deviation){
        printf("ERROR: Detected in checks in linepolarization.c\n");
        printf("ERROR: B is not a unit vector.\n");
        printf("ERROR: abs(b) = %e\n", absB);
        exit(1);
    }
    
    double absn;
    absn = sqrt(n[0]*n[0] + n[1]*n[1] + n[2]*n[2]);
    if(fabs(absn-1.) > deviation){
        printf("ERROR: Detected in checks in linepolarization.c\n");
        printf("ERROR: n is not a unit vector.\n");
        printf("ERROR: abs(n) = %e\n", absn);
        printf("ERROR: n[0] = %e\n", n[0]);
        printf("ERROR: n[1] = %e\n", n[1]);
        printf("ERROR: n[2] = %e\n", n[2]);
        exit(1);
    }
    
    double abse_perp;
    abse_perp = sqrt(e_perp[0]*e_perp[0] + e_perp[1]*e_perp[1] + e_perp[2]*e_perp[2]);
    if(fabs(abse_perp-1.) > deviation){
        printf("ERROR: Detected in checks in linepolarization.c\n");
        printf("ERROR: e_perp is not a unit vector.\n");
        printf("ERROR: abs(e_perp) = %e\n", abse_perp);
        exit(1);
    }
    
    double abse_para;
    abse_para = sqrt(e_para[0]*e_para[0] + e_para[1]*e_para[1] + e_para[2]*e_para[2]);
    if(fabs(abse_para-1.) > deviation){
        printf("ERROR: Detected in checks in linepolarization.c\n");
        printf("ERROR: e_para is not a unit vector.\n");
        printf("ERROR: abs(e_para) = %e\n", abse_para);
        exit(1);
    }
}



void linepolarization_PolarizationFrame2CartesianFrame_snu_and_tau_simultaneously(struct grid *g, int here, double *n, double snu_para, double snu_perp, double dtau_para, double dtau_perp, double *snu_pol, double *dtau_pol){
    
    double B[3], absB;
    double e_perp[3], e_para[3];
    double Bcrossn[3], absBcrossn, Bdotn;
    double deviation = 1e-14;
    
    // get magnetic field:
    magneticfield(g[here].x[0], g[here].x[1], g[here].x[2], B);
    
    // normalize the magnetic field vector:
    absB = sqrt(B[0]*B[0] + B[1]*B[1] + B[2]*B[2]);
    B[0] /= absB;
    B[1] /= absB;
    B[2] /= absB;
    
    // compute polarization vectors:
    //  They are perpendicular and parallel to the projection of the magnetic field vector B[]
    //  in the plane perpendicular to the propagation direction of the light n[].
    Bdotn = B[0]*n[0] + B[1]*n[1] + B[2]*n[2];
    Bcrossn[0] = B[1]*n[2] - B[2]*n[1];
    Bcrossn[1] = B[2]*n[0] - B[0]*n[2];
    Bcrossn[2] = B[0]*n[1] - B[1]*n[0];
    absBcrossn = sqrt(Bcrossn[0]*Bcrossn[0] + Bcrossn[1]*Bcrossn[1] + Bcrossn[2]*Bcrossn[2]);
    e_perp[0] = Bcrossn[0] / absBcrossn;
    e_perp[1] = Bcrossn[1] / absBcrossn;
    e_perp[2] = Bcrossn[2] / absBcrossn;
    e_para[0] = (B[0] - n[0]*Bdotn) / absBcrossn;
    e_para[1] = (B[1] - n[1]*Bdotn) / absBcrossn;
    e_para[2] = (B[2] - n[2]*Bdotn) / absBcrossn;
    
    double e_perpdote_para;
    e_perpdote_para = e_perp[0]*e_para[0] + e_perp[1]*e_para[1] + e_perp[2]*e_para[2];
    if(fabs(e_perpdote_para) > deviation){
        printf("ERROR: Detected in checks in linepolarization.c\n");
        printf("ERROR: e_perp is not perpendicular to e_para.\n");
        printf("ERROR: e_perp dot e_para = %e\n", e_perpdote_para);
        exit(1);
    }
    
    e_perp[0] = fabs(e_perp[0]);
    e_perp[1] = fabs(e_perp[1]);
    e_perp[2] = fabs(e_perp[2]);
    e_para[0] = fabs(e_para[0]);
    e_para[1] = fabs(e_para[1]);
    e_para[2] = fabs(e_para[2]);
    
    double e_para_tot, e_perp_tot;
    e_para_tot = e_para[0] + e_para[1] + e_para[2];
    e_perp_tot = e_perp[0] + e_perp[1] + e_perp[2];
    
    snu_pol[0]  = e_para[0]/e_para_tot * snu_para  + e_perp[0]/e_perp_tot * snu_perp;
    snu_pol[1]  = e_para[1]/e_para_tot * snu_para  + e_perp[1]/e_perp_tot * snu_perp;
    snu_pol[2]  = e_para[2]/e_para_tot * snu_para  + e_perp[2]/e_perp_tot * snu_perp;
    dtau_pol[0] = e_para[0]/e_para_tot * dtau_para + e_perp[0]/e_perp_tot * dtau_perp;
    dtau_pol[1] = e_para[1]/e_para_tot * dtau_para + e_perp[1]/e_perp_tot * dtau_perp;
    dtau_pol[2] = e_para[2]/e_para_tot * dtau_para + e_perp[2]/e_perp_tot * dtau_perp;
    
    
    // special handling, if the light travels exactly along the magnetic field lines:
    if(absBcrossn < 1e-14){
        if(fabs(snu_para - snu_perp)/snu_para > 1e-14){
            printf("ERROR: Detected in PolarizationFrame2CartesianFrame() in linepolarization.c\n");
            printf("ERROR: The light is moving along the magnetic field lines, but snu_para != snu_perp.\n");
            printf("ERROR: fabs(snu_para - snu_perp)/snu_para = %e\n", fabs(snu_para - snu_perp)/snu_para);
            exit(1);
        }
        if(fabs(dtau_para - dtau_perp)/dtau_para > 1e-14){
            printf("ERROR: Detected in PolarizationFrame2CartesianFrame() in linepolarization.c\n");
            printf("ERROR: The light is moving along the magnetic field lines, but dtau_para != dtau_perp.\n");
            printf("ERROR: fabs(dtau_para - dtau_perp)/dtau_para = %e\n", fabs(dtau_para - dtau_perp)/dtau_para);
            exit(1);
        }
        snu_pol[0]  = (n[1]+n[2]) / (n[0]+n[1]+n[2]) * 0.5 * (snu_para + snu_perp);
        snu_pol[1]  = (n[2]+n[0]) / (n[0]+n[1]+n[2]) * 0.5 * (snu_para + snu_perp);
        snu_pol[2]  = (n[0]+n[1]) / (n[0]+n[1]+n[2]) * 0.5 * (snu_para + snu_perp);
        dtau_pol[0] = (n[1]+n[2]) / (n[0]+n[1]+n[2]) * 0.5 * (dtau_para + dtau_perp);
        dtau_pol[1] = (n[2]+n[0]) / (n[0]+n[1]+n[2]) * 0.5 * (dtau_para + dtau_perp);
        dtau_pol[2] = (n[0]+n[1]) / (n[0]+n[1]+n[2]) * 0.5 * (dtau_para + dtau_perp);
    }
    
    // checks
    absB = sqrt(B[0]*B[0] + B[1]*B[1] + B[2]*B[2]);
    if(fabs(absB-1.) > deviation){
        printf("ERROR: Detected in checks in linepolarization.c\n");
        printf("ERROR: B is not a unit vector.\n");
        printf("ERROR: abs(b) = %e\n", absB);
        exit(1);
    }
    
    double absn;
    absn = sqrt(n[0]*n[0] + n[1]*n[1] + n[2]*n[2]);
    if(fabs(absn-1.) > deviation){
        printf("ERROR: Detected in checks in linepolarization.c\n");
        printf("ERROR: n is not a unit vector.\n");
        printf("ERROR: abs(n) = %e\n", absn);
        printf("ERROR: n[0] = %e\n", n[0]);
        printf("ERROR: n[1] = %e\n", n[1]);
        printf("ERROR: n[2] = %e\n", n[2]);
        exit(1);
    }
    
    double abse_perp;
    abse_perp = sqrt(e_perp[0]*e_perp[0] + e_perp[1]*e_perp[1] + e_perp[2]*e_perp[2]);
    if(fabs(abse_perp-1.) > deviation){
        printf("ERROR: Detected in checks in linepolarization.c\n");
        printf("ERROR: e_perp is not a unit vector.\n");
        printf("ERROR: abs(e_perp) = %e\n", abse_perp);
        exit(1);
    }
    
    double abse_para;
    abse_para = sqrt(e_para[0]*e_para[0] + e_para[1]*e_para[1] + e_para[2]*e_para[2]);
    if(fabs(abse_para-1.) > deviation){
        printf("ERROR: Detected in checks in linepolarization.c\n");
        printf("ERROR: e_para is not a unit vector.\n");
        printf("ERROR: abs(e_para) = %e\n", abse_para);
        exit(1);
    }
    
    if(fabs(((snu_pol[0] + snu_pol[1] + snu_pol[2]) - (snu_para + snu_perp))/(snu_para + snu_perp)) > deviation){
        printf("ERROR: Detected in checks in linepolarization.c\n");
        printf("ERROR: snu_pol[0] + snu_pol[1] + snu_pol[2] != snu_para + snu_perp.\n");
        printf("ERROR: ((snu_pol[0] + snu_pol[1] + snu_pol[2]) - (snu_para + snu_perp))/(snu_para + snu_perp) = %e\n", ((snu_pol[0] + snu_pol[1] + snu_pol[2]) - (snu_para + snu_perp))/(snu_para + snu_perp));
        exit(1);
    }
}



void linepolarization_CartesianFrame2PolarizationFrame(struct grid *g, int here, double *n, double *Icartesian, double *Ipara, double *Iperp){
    
    double B[3], Bdotn;
    double Bprojected[3], absBprojected;
    double BprojecteddotIcartesian;
    
    // get magnetic field:
    magneticfield(g[here].x[0], g[here].x[1], g[here].x[2], B);
    
    // compute B projected:
    Bdotn = B[0]*n[0] + B[1]*n[1] + B[2]*n[2];
    
    Bprojected[0] = B[0] - n[0]*Bdotn;
    Bprojected[1] = B[1] - n[1]*Bdotn;
    Bprojected[2] = B[2] - n[2]*Bdotn;
    
    // normalize the projected magnetic field vector:
    absBprojected = sqrt(Bprojected[0]*Bprojected[0] + Bprojected[1]*Bprojected[1] + Bprojected[2]*Bprojected[2]);
    Bprojected[0] /= absBprojected;
    Bprojected[1] /= absBprojected;
    Bprojected[2] /= absBprojected;
    
    // vector wise:
    BprojecteddotIcartesian = Icartesian[0]*Bprojected[0] + Icartesian[1]*Bprojected[1] + Icartesian[2]*Bprojected[2];
    
    *Ipara = BprojecteddotIcartesian;
    *Iperp = sqrt(pow(Icartesian[0] - *Ipara*Bprojected[0], 2) + pow(Icartesian[1] - *Ipara*Bprojected[1], 2) + pow(Icartesian[2] - *Ipara*Bprojected[2], 2));
}



void linepolarization_CartesianFrame2StokesVector(image *img, int im, double *Icartesian, double *Stokes){
    
    double Ix, Iy, Iplus, Iminus, Iright, Ileft;
    double ex[3], ey[3], eplus[3], eminus[3];
    double lineofsight[3], Iparalineofsight[3], Iperplineofsight[3];
    double theta, phi;
    
    linepolarization_GetPropagationDirection_raytrace(lineofsight, img, im);
    
    // normalize line of sight vector:
    vectoralgebra_Normalize(lineofsight);
    
    // project Icartesian along line of sight:
    vectoralgebra_Project(Icartesian, lineofsight, Iparalineofsight);
    
    // compute part of Icartesian perpendicular to line of sight:
    Iperplineofsight[0] = Icartesian[0] - Iparalineofsight[0];
    Iperplineofsight[1] = Icartesian[1] - Iparalineofsight[1];
    Iperplineofsight[2] = Icartesian[2] - Iparalineofsight[2];
    
    //
    // Linear Polarization:
    //
    // Linear Polarization, Horizontal (90°) and Vertical (0°):
    //
    // compute coordinate system (unit vectors):
    theta = img[im].theta;
    phi   = img[im].phi;
    
    ex[0] = cos(phi);
    ex[1] = sin(phi)*cos(theta);
    ex[2] = sin(phi)*sin(theta);
    
    // ey = ez cross ex and ez = line of sight:
    vectoralgebra_CrossProduct(lineofsight, ex, ey);
    
    // compute polarized Intensities:
    Ix = vectoralgebra_DotProduct(Iperplineofsight, ex);
    Iy = vectoralgebra_DotProduct(Iperplineofsight, ey);
    
    //
    // Linear Polarization, +/- 45°:
    //
    // compute coordinate system (unit vectors):
    eplus[0] = 1.0/sqrt(2) * (ex[0]+ey[0]);
    eplus[1] = 1.0/sqrt(2) * (ex[1]+ey[1]);
    eplus[2] = 1.0/sqrt(2) * (ex[2]+ey[2]);
    
    eminus[0] = 1.0/sqrt(2) * (ey[0]-ex[0]);
    eminus[1] = 1.0/sqrt(2) * (ey[1]-ex[1]);
    eminus[2] = 1.0/sqrt(2) * (ey[2]-ex[2]);
    
    // compute polarized Intensities:
    Iplus  = vectoralgebra_DotProduct(Iperplineofsight, eplus );
    Iminus = vectoralgebra_DotProduct(Iperplineofsight, eminus);
    
    //
    // Circular Polarization:
    //  circular polarization not inplemented yet.
    //
    Iright = 0;
    Ileft = Iright;
    
    //
    // Write Stokes Vector:
    //
    //  Stokes[0] = I
    //  Stokes[1] = Q
    //  Stokes[2] = U
    //  Stokes[3] = V
    //
    Stokes[0] = Ix+Iy;
    Stokes[1] = Ix-Iy;
    Stokes[2] = Iplus-Iminus;
    Stokes[3] = Iright-Ileft;
    
    // checks
    if(!vectoralgebra_CheckUnitVector(ex)){
        printf("ERROR: Detected in linepolarization_CartesianFrame2StokesVector() in linepolarization.c\n");
        printf("ERROR: ex is not an unit vector.\n");
        exit(1);
    }
    if(!vectoralgebra_CheckUnitVector(ey)){
        printf("ERROR: Detected in linepolarization_CartesianFrame2StokesVector() in linepolarization.c\n");
        printf("ERROR: ey is not an unit vector.\n");
        exit(1);
    }
    if(!vectoralgebra_CheckUnitVector(eplus)){
        printf("ERROR: Detected in linepolarization_CartesianFrame2StokesVector() in linepolarization.c\n");
        printf("ERROR: eplus is not an unit vector.\n");
        exit(1);
    }
    if(!vectoralgebra_CheckUnitVector(eminus)){
        printf("ERROR: Detected in linepolarization_CartesianFrame2StokesVector() in linepolarization.c\n");
        printf("ERROR: eminus is not an unit vector.\n");
        exit(1);
    }
    if(!vectoralgebra_CheckUnitVector(lineofsight)){
        printf("ERROR: Detected in linepolarization_CartesianFrame2StokesVector() in linepolarization.c\n");
        printf("ERROR: lineofsight is not an unit vector.\n");
        exit(1);
    }
    if(!vectoralgebra_CheckPerpendicularity(ex, ey)){
        printf("ERROR: Detected in linepolarization_CartesianFrame2StokesVector() in linepolarization.c\n");
        printf("ERROR: ex is not perpendicular to ey.\n");
        exit(1);
    }
    if(!vectoralgebra_CheckPerpendicularity(eplus, eminus)){
        printf("ERROR: Detected in linepolarization_CartesianFrame2StokesVector() in linepolarization.c\n");
        printf("ERROR: eplus is not perpendicular to eminus.\n");
        exit(1);
    }
    if(!vectoralgebra_CheckPerpendicularity(ex, lineofsight)){
        printf("ERROR: Detected in linepolarization_CartesianFrame2StokesVector() in linepolarization.c\n");
        printf("ERROR: ex is not perpendicular to lineofsight.\n");
        exit(1);
    }
    if(!vectoralgebra_CheckPerpendicularity(ey, lineofsight)){
        printf("ERROR: Detected in linepolarization_CartesianFrame2StokesVector() in linepolarization.c\n");
        printf("ERROR: ey is not perpendicular to lineofsight.\n");
        exit(1);
    }
    if(!vectoralgebra_CheckPerpendicularity(eplus, lineofsight)){
        printf("ERROR: Detected in linepolarization_CartesianFrame2StokesVector() in linepolarization.c\n");
        printf("ERROR: eplus is not perpendicular to lineofsight.\n");
        exit(1);
    }
    if(!vectoralgebra_CheckPerpendicularity(eminus, lineofsight)){
        printf("ERROR: Detected in linepolarization_CartesianFrame2StokesVector() in linepolarization.c\n");
        printf("ERROR: eminus is not perpendicular to lineofsight.\n");
        exit(1);
    }
    if(Ix < 0.0){
        printf("ERROR: Detected in linepolarization_CartesianFrame2StokesVector() in linepolarization.c\n");
        printf("ERROR: Ix is negative.\n");
        exit(1);
    }
    if(Iy < 0.0){
        printf("ERROR: Detected in linepolarization_CartesianFrame2StokesVector() in linepolarization.c\n");
        printf("ERROR: Iy is negative.\n");
        exit(1);
    }
    if(Iplus < 0.0){
        printf("ERROR: Detected in linepolarization_CartesianFrame2StokesVector() in linepolarization.c\n");
        printf("ERROR: Iplus is negative.\n");
        exit(1);
    }
    if(Iminus < 0.0){
        printf("ERROR: Detected in linepolarization_CartesianFrame2StokesVector() in linepolarization.c\n");
        printf("ERROR: Iminus is negative.\n");
        exit(1);
    }
    if(!(pow(Stokes[1], 2) + pow(Stokes[2], 2) + pow(Stokes[3], 2) <= pow(Stokes[0], 2))){
        printf("ERROR: Detected in linepolarization_CartesianFrame2StokesVector() in linepolarization.c\n");
        printf("ERROR: Polarization degree is larger than unity.\n");
        exit(1);
    }
}


// currently not needed, only used for printing the ZeemanLinewidth()
double Lande(int J, int S, int L){
    return 1.0;
    /*return
     1.0
     +
     (
     J * (J + 1.0)
     +
     S * (S + 1.0)
     -
     L * (L + 1.0)
     )
     /
     (
     2.0 * J * (J + 1.0)
     );*/
    //=> results in nan for J = 0
}

double DopplerLinewidth(double x, double y, double z, molData *m, int iline, int i){
    double dopplerb;
    doppler(x, y, z, &dopplerb);
    return dopplerb/CLIGHT * m[i].freq[iline];
}

double ZeemanLinewidth(double magneticfield){
    return Lande(1,0,1) * BOHRMAG * magneticfield / HPLANCK;
}

double NaturalLinewidth(molData *m, int iline, int i){
    return m[i].aeinst[iline]/(4.*PI);
}

#endif
