/*
 *  aux.c
 *  LIME, The versatile 3D line modeling environment 
 *
 *  Created by Christian Brinch on 16/11/06.
 *  Copyright 2006-2011, Christian Brinch, 
 *  <brinch@strw.leidenuniv.nl>
 *  Sterrewacht Leiden
 *  Leiden University
 *	All rights reserved.
 *
 */

#include "lime.h"
#include <gsl/gsl_sort.h>
#include <gsl/gsl_statistics.h>

int nblendedlines;


void 
parseInput(inputPars *par, image **img, molData **m){
	int i,id;

	/* Set default values */
	par->tcmb = 2.728;
	par->lte_only=0;
	par->sampling=0;
	par->blend=0;

	/* Allocate space for output fits images */
	(*img)=malloc(sizeof(image)*100);
	par->moldatfile=malloc(sizeof(char *) * 100);
	for(id=0;id<100;id++){
		(*img)[id].filename=NULL;
		par->moldatfile[id]=NULL;
	}
	input(par, *img);
	id=0;
	while((*img)[++id].filename!=NULL);
	par->nImages=id;
	if(par->nImages==0) {
		if(!silent) bail_out("Error: no images defined");
		exit(1);
	}

	free(*img);
	*img=malloc(sizeof(image)*par->nImages);

	id=-1;
	while(par->moldatfile[++id]!=NULL);
	par->nSpecies=id;

	free(par->moldatfile);
	par->moldatfile=malloc(sizeof(char *)*par->nSpecies);

	/* Set defaults and read inputPars and img[] */
	for(i=0;i<par->nImages;i++) {
	  (*img)[i].source_vel=0.0;
	  (*img)[i].phi=0.0;
	  (*img)[i].nchan=0;
	  (*img)[i].velres=-1.;
	  (*img)[i].trans=-1;
	  (*img)[i].freq=-1.;
	  (*img)[i].bandwidth=-1.;
	}
	input(par,*img);
	par->ncell=par->pIntensity+par->sinkPoints;


	/* Allocate pixel space and parse image information */
	for(i=0;i<par->nImages;i++){
		(*img)[i].imgres=(*img)[i].imgres/206264.806;
		(*img)[i].pixel = malloc(sizeof(spec)*(*img)[i].pxls*(*img)[i].pxls);
		for(id=0;id<((*img)[i].pxls*(*img)[i].pxls);id++){
			(*img)[i].pixel[id].intense = malloc(sizeof(double)*(*img)[i].nchan);
			(*img)[i].pixel[id].tau = malloc(sizeof(double)*(*img)[i].nchan);
#if LINEPOLARIZATION == ON
			(*img)[i].pixel[id].intensepx = malloc(sizeof(double)*(*img)[i].nchan);
			(*img)[i].pixel[id].intensepy = malloc(sizeof(double)*(*img)[i].nchan);
			(*img)[i].pixel[id].intensepz = malloc(sizeof(double)*(*img)[i].nchan);
			(*img)[i].pixel[id].taupx     = malloc(sizeof(double)*(*img)[i].nchan);
			(*img)[i].pixel[id].taupy     = malloc(sizeof(double)*(*img)[i].nchan);
			(*img)[i].pixel[id].taupz     = malloc(sizeof(double)*(*img)[i].nchan);
			// TODO: This memory seems not to be freed anywhere, also not in LIME itself (without line polarization)?
#endif
		}
		
		if((*img)[i].nchan == 0 && (*img)[i].velres<0 ){
			/* Assume continuum image */
			(*img)[i].nchan=1;
			if(par->nSpecies!=0 || (*img)[i].trans>-1 || (*img)[i].bandwidth>-1. || (*img)[i].freq==0 || par->dust==NULL){
				if(!silent) bail_out("Error: Image keywords are ambiguous");
				exit(1);
			}
			(*img)[i].doline=0;		  
		} else if (((*img)[i].nchan>0 || (*img)[i].velres > 0)){
			/* Assume line image */	
			if(par->moldatfile==NULL){
				if(!silent) bail_out("Error: No data file is specified for line image.");
				exit(1);		
			}
			if(((*img)[i].trans>-1 && (*img)[i].freq>-1) || ((*img)[i].trans<0 && (*img)[i].freq<0)){
				if(!silent) bail_out("Error: Specify either frequency or transition ");
				exit(1);						
			}	
			if(((*img)[i].nchan==0 && (*img)[i].bandwidth<0) || ((*img)[i].bandwidth<0 && (*img)[i].velres<0)){
				if(!silent) bail_out("Error: Image keywords are not set properly");
				exit(1);
			}
		 	(*img)[i].doline=1;
		}
	}
	/* Allocate moldata array */
	(*m)=malloc(sizeof(molData)*par->nSpecies);
}

float 
invSqrt(float x){
	/* The magic Quake(TM) fast inverse square root algorithm   */
	/* Can _only_ be used on 32-bit machine architectures       */
	float xhalf = 0.5f*x;
    int i = *(int*)&x;
    i = 0x5f3759df - (i>>1);
    x = *(float*)&i;
    x = x*(1.5f - xhalf*x*x);
    return x;
}

void 
continuumSetup(int im, image *img, molData *m, inputPars *par, struct grid *g){
	int id;
	img[im].nchan=1;
	img[im].trans=0;
	m[0].nline=1;
	m[0].freq= malloc(sizeof(double)*m[0].nline);
	m[0].freq[0]=img[im].freq;
	for(id=0;id<par->ncell;id++) {
	  g[id].mol=malloc(sizeof(struct populations)*1);	
	  g[id].mol[0].dust = malloc(sizeof(double)*m[0].nline);
	  g[id].mol[0].knu  = malloc(sizeof(double)*m[0].nline);
	} 
	kappa(m,g,par,0);	
}

void
lineCount(int n,molData *m,int **counta,int **countb,int *nlinetot){
	int ispec,iline,count;	
	
	*nlinetot=0;
	for(ispec=0;ispec<n;ispec++) *nlinetot+=m[ispec].nline;
	*counta=malloc(sizeof(int)* *nlinetot);
	*countb=malloc(sizeof(int)* *nlinetot);
	count=0;
	for(ispec=0;ispec<n;ispec++) {
		for(iline=0;iline<m[ispec].nline;iline++){
 			(*counta)[count]=ispec;
			(*countb)[count++]=iline;			
		}
	}
}

void
lineBlend(molData *m, inputPars *par, blend **matrix){
	int iline, jline, nlinetot=0,c;
	int *counta,*countb;

	lineCount(par->nSpecies, m, &counta, &countb, &nlinetot);

	c=0;
	for(iline=0;iline<nlinetot;iline++){
	  for(jline=0;jline<nlinetot;jline++){  
		if(fabs((m[counta[jline]].freq[countb[jline]]-m[counta[iline]].freq[countb[iline]])/m[counta[iline]].freq[countb[iline]]*CLIGHT) < blendmask
			&& iline !=jline) c++;
		}
	}
	
	nblendedlines = c;
	
	if(c>0){
	  if(par->blend){
			if(!silent) warning("There are blended lines (Line blending is switched on)");
		} else {
			if(!silent) warning("There are blended lines (Line blending is switched off)");
	  }
		//printf("\nblend: c = %d\n", c); bail_out("STOP");exit(1);
	  (*matrix)=malloc(sizeof(blend)*c);

	  c=0;
	  for(iline=0;iline<nlinetot;iline++){
   	    for(jline=0;jline<nlinetot;jline++){  
  		  if(fabs((m[counta[jline]].freq[countb[jline]]-m[counta[iline]].freq[countb[iline]])/m[counta[iline]].freq[countb[iline]]*CLIGHT) < blendmask
			&& iline != jline){
		    (*matrix)[c].line1=iline;
		    (*matrix)[c].line2=jline;
			  //printf("\n");
			  //printf("c =     %d\n", c);
			  //printf("iline = %d\n", iline);
			  //printf("jline = %d\n", jline);
			  //printf("matrix[c].line1 = %d\n", (*matrix)[c].line1);
			  //printf("matrix[c].line2 = %d\n", (*matrix)[c].line2);
		    (*matrix)[c++].deltav=-(m[counta[jline]].freq[countb[jline]]-m[counta[iline]].freq[countb[iline]])/m[counta[iline]].freq[countb[iline]]*CLIGHT;				
		  }			
	    }						
	  }		
	}
	free(counta);
	free(countb);
	//bail_out("STOP");exit(1);
}

void 
levelPops(molData *m, inputPars *par, struct grid *g){

	int id,conv=0,iter,ilev,prog=0,ispec,c=0;
	double percent=0.,pstate,*median,result1=0,result2=0;
	blend *matrix;
	struct statistics { double *pop, *ave, *sigma; } *stat;
	
	stat=malloc(sizeof(struct statistics)*par->pIntensity);

	for(id=0;id<par->ncell;id++) {
	  g[id].mol=malloc(sizeof(struct populations)*par->nSpecies);	
	} 	

	/* Random number generator */
	const gsl_rng *ran = gsl_rng_alloc(gsl_rng_ranlxs2);	
	gsl_rng_set(ran,time(0));
	
	/* Read in all molecular data */
	for(id=0;id<par->nSpecies;id++) molinit(m,par,g,id);
	/* Check for blended lines */
	lineBlend(m,par,&matrix);

	if(par->lte_only) LTE(par,g,m);

	for(id=0;id<par->pIntensity;id++){
	  stat[id].pop=malloc(sizeof(double)*m[0].nlev*5);
	  stat[id].ave=malloc(sizeof(double)*m[0].nlev);
	  stat[id].sigma=malloc(sizeof(double)*m[0].nlev);
	  for(ilev=0;ilev<m[0].nlev;ilev++) {
	    for(iter=0;iter<5;iter++) stat[id].pop[ilev+m[0].nlev*iter]=g[id].mol[0].pops[ilev];
	  }
	}

	if(par->lte_only==0){
	  do{
		if(!silent) progressbar2(prog++, 0, result1, result2);
		pstate=0.;
//		printf(" conv = %d\n", conv);//exit(1);
//		printf(" par->ncell = %d\n", par->ncell);//exit(1);
		for(id=0;id<par->ncell && !g[id].sink;id++){
			if(!silent) progressbar((double)id/par->pIntensity,10);
			
//			printf("  id = %d\n", id);//exit(1);
			
			for(ilev=0;ilev<m[0].nlev;ilev++) {
				for(iter=0;iter<4;iter++) stat[id].pop[ilev+m[0].nlev*iter]=stat[id].pop[ilev+m[0].nlev*(iter+1)];
				stat[id].pop[ilev+m[0].nlev*4]=g[id].mol[0].pops[ilev];
			}
				
			if(g[id].dens[0] > 0 && g[id].t[0] > 0){
			  photon(id,g,m,0,ran,par,matrix);
			  for(ispec=0;ispec<par->nSpecies;ispec++) stateq(id,g,m,&pstate,ispec,par);
			}
				
			for(ilev=0;ilev<m[0].nlev;ilev++) {
				stat[id].ave[ilev]=0;
				for(iter=0;iter<5;iter++) stat[id].ave[ilev]+=stat[id].pop[ilev+m[0].nlev*iter];
				stat[id].ave[ilev]=stat[id].ave[ilev]/5.;
				stat[id].sigma[ilev]=0;
				for(iter=0;iter<5;iter++) stat[id].sigma[ilev]+=pow(stat[id].pop[ilev+m[0].nlev*iter]-stat[id].ave[ilev],2);
				stat[id].sigma[ilev]=sqrt(stat[id].sigma[ilev])/5.;
				if(g[id].mol[0].pops[ilev] > 1e-12) c++;
			}
		}
//		if(conv == 1) exit(1);
			
		median=malloc(sizeof(double)*c);
		c=0;
		for(id=0;id<par->pIntensity;id++){
		  for(ilev=0;ilev<m[0].nlev;ilev++){
			if(g[id].mol[0].pops[ilev] > 1e-12) median[c++]=g[id].mol[0].pops[ilev]/stat[id].sigma[ilev];
		  }
		}

		gsl_sort(median, 1, c);
		if(conv>1){
		  result1=median[0];
		  result2 =gsl_stats_median_from_sorted_data(median, 1, c);
		}
		free(median);

		if(!silent) progressbar2(prog, percent, result1, result2);		
		if(par->outputfile) popsout(par,g,m);
		  
#if DEBUG_SINGLE_ITERATION == YES
		//bail_out("STOP HERE");
		//exit(1);
#endif
		
	  } while(conv++<NITERATIONS);
	}

}

	  	

