/*
 *  stateq.c
 *  LIME, The versatile 3D line modeling environment 
 *
 *  Created by Christian Brinch on 15/11/06.
 *  Copyright 2006-2011, Christian Brinch, 
 *  <brinch@strw.leidenuniv.nl>
 *  Sterrewacht Leiden
 *  Leiden University 
 *	All rights reserved.
 *
 */

#include "lime.h"
#include <gsl/gsl_vector.h>
#include <gsl/gsl_permutation.h>
#if LINEPOLARIZATION == ON
#include "linepolarization.h"
#endif


void
stateq(int id, struct grid *g, molData *m, double *pstate, int ispec, inputPars *par){
	int t,s,iter;
	double *opop, *oopop;
	double diff;

	// Allocate memory:
	gsl_matrix *matrix = gsl_matrix_alloc(m[ispec].nlev+1, m[ispec].nlev+1);
	gsl_matrix *reduc  = gsl_matrix_alloc(m[ispec].nlev, m[ispec].nlev);
	gsl_vector *newpop = gsl_vector_alloc(m[ispec].nlev);
	gsl_vector *oldpop = gsl_vector_alloc(m[ispec].nlev);
	gsl_permutation *p = gsl_permutation_alloc (m[ispec].nlev);

	opop	 = malloc(sizeof(double)*m[ispec].nlev);	
	oopop	 = malloc(sizeof(double)*m[ispec].nlev);

	// Set starting values of vectors:
	for(t=0;t<m[ispec].nlev-1;t++){
		opop[t]=0.;
		oopop[t]=0.;
		gsl_vector_set(oldpop,t,0.);
	}
	gsl_vector_set(oldpop,m[ispec].nlev-1,1.);
	diff=1;
	iter=0;

	// Main loop: Iterative solver
	while((diff>TOL && iter<MAXITER) || iter<5){
        getjbar(id,m,g,par);
	    getmatrix(id,matrix,m,g,ispec); // see routine 'getmatrix()' below
	    // Set entries of reduced matrix for system of equations:
		for(s=0;s<m[ispec].nlev;s++){
			for(t=0;t<m[ispec].nlev-1;t++){
				gsl_matrix_set(reduc,t,s,gsl_matrix_get(matrix,t,s));
			}
			gsl_matrix_set(reduc,m[ispec].nlev-1,s,1.);
		}
		
		// Solve the implicit problem:
		gsl_linalg_LU_decomp(reduc,p,&s);
		gsl_linalg_LU_solve(reduc,p,oldpop,newpop);

		// Compute convergence criterion:
		diff=0.;
		for(t=0;t<m[ispec].nlev;t++){
			gsl_vector_set(newpop,t,gsl_max(gsl_vector_get(newpop,t),1e-30));
			oopop[t]=opop[t];
			opop[t]=g[id].mol[ispec].pops[t];
			g[id].mol[ispec].pops[t]=gsl_vector_get(newpop,t);
			//printf("t =                        %d\n", t);
			//printf("g[id].mol[ispec].pops[t] = %e\n", g[id].mol[ispec].pops[t]);
			//exit(1);
			if(gsl_min(g[id].mol[ispec].pops[t],gsl_min(opop[t],oopop[t]))>minpop){
				diff=gsl_max(fabs(g[id].mol[ispec].pops[t]-opop[t])/g[id].mol[ispec].pops[t],gsl_max(fabs(g[id].mol[ispec].pops[t]-oopop[t])/g[id].mol[ispec].pops[t],diff));
			}
		}
		iter++;
	}

	// Check for convergence:
	if(diff>TOL) *pstate=diff;

	// Free memory:
	gsl_matrix_free(matrix);
	gsl_matrix_free(reduc);	
	gsl_vector_free(oldpop);
	gsl_vector_free(newpop);
	gsl_permutation_free(p);
	free(opop);
	free(oopop);
}


void 
getmatrix(int id, gsl_matrix *matrix, molData *m, struct grid *g, int ispec)
{
	int p,t,k,l,ipart;
	struct getmatrix {
		double *ctot;
		gsl_matrix * colli;
	} *partner;

#if LINEPOLARIZATION == ON
	int Jup, Jdown, Mup, Mdown, dM;
#endif
	double jbar;
	
	// Initialize matrices:
	partner= malloc(sizeof(struct getmatrix)*m[ispec].npart);
	
	for(ipart=0;ipart<m[ispec].npart;ipart++){
		partner[ipart].colli = gsl_matrix_alloc(m[ispec].nlev+1,m[ispec].nlev+1);
		partner[ipart].ctot  = malloc(sizeof(double)*m[ispec].nlev);
		for(t=0;t<m[ispec].nlev+1;t++){
			for(p=0;p<m[ispec].nlev+1;p++){		
				gsl_matrix_set(matrix, t, p, 0.); 
				gsl_matrix_set(partner[ipart].colli, t, p, 0.); 
			}
		}
	}


	// Set radiative transitions of total matrix:
	for(t=0;t<m[ispec].nline;t++){
#if LINEPOLARIZATION == ON
		// Compute the energy level number from the l and m of the upper (k) and lower (l) level of the radiative transition:
		Jup   = m[ispec].lau[t];
		Jdown = m[ispec].lal[t];
		Mup   = m[ispec].mau[t];
		Mdown = m[ispec].mal[t];
		dM = abs(Mup-Mdown);
		
		k = linepolarization_GetEnergyLevelNumber(  Jup,   Mup, m, ispec);
		l = linepolarization_GetEnergyLevelNumber(Jdown, Mdown, m, ispec);
		
		// Compute the mean radiation field:
		if(linepolarization_AreMagneticSublevelsTreatedExplicitly(Jup) && linepolarization_AreMagneticSublevelsTreatedExplicitly(Jdown)){
			if(dM == 0){
				jbar = m[ispec].jbar_dM0[t];
			}
			else if(dM == 1){
				jbar = m[ispec].jbar_dM1[t];
			}
			else{
				bail_out("ERROR: dM not equal to 1 or 0.");
			}
			jbar *= 2.0*Jup+1.0;
			jbar /= 2.0*Jdown+1.0;
		}
		else if(!(linepolarization_AreMagneticSublevelsTreatedExplicitly(Jup) || linepolarization_AreMagneticSublevelsTreatedExplicitly(Jdown))){
			jbar = m[ispec].jbar_dM0[t] + 2.0*m[ispec].jbar_dM1[t];
			jbar *= 2.0;
		}
		else{
			jbar = m[ispec].jbar_para[t] + m[ispec].jbar_perp[t];
		}
#else
		k = m[ispec].lau[t];
		l = m[ispec].lal[t];
		jbar = m[ispec].jbar[t];
#endif

		gsl_matrix_set(matrix, k, k, gsl_matrix_get(matrix, k, k)+m[ispec].beinstu[t]*jbar+m[ispec].aeinst[t]);
		gsl_matrix_set(matrix, l, l, gsl_matrix_get(matrix, l, l)+m[ispec].beinstl[t]*jbar);
		gsl_matrix_set(matrix, k, l, gsl_matrix_get(matrix, k, l)-m[ispec].beinstl[t]*jbar);
		gsl_matrix_set(matrix, l, k, gsl_matrix_get(matrix, l, k)-m[ispec].beinstu[t]*jbar-m[ispec].aeinst[t]);
	}

    // Set entries of collision matrix:
	for(ipart=0;ipart<m[ispec].npart;ipart++){
		for(t=0;t<m[ispec].ntrans[ipart];t++){
#if LINEPOLARIZATION == ON
			// Compute the energy level number from the l (and m) of the upper (k) and lower (l) level of the collision transition:
			Jup   = m[ispec].lcu[t];
			Jdown = m[ispec].lcl[t];
			Mup   = m[ispec].mcu[t];
			Mdown = m[ispec].mcl[t];

			k = linepolarization_GetEnergyLevelNumber(  Jup,   Mup, m, ispec);
			l = linepolarization_GetEnergyLevelNumber(Jdown, Mdown, m, ispec);
#else
			k = m[ispec].lcu[t];
			l = m[ispec].lcl[t];
#endif
			gsl_matrix_set(partner[ipart].colli, k, l, g[id].mol[ispec].partner[ipart].down[t]);
			gsl_matrix_set(partner[ipart].colli, l, k, g[id].mol[ispec].partner[ipart].up[t]);
		}
		// Sum the rows of the collision matrix (excluding the last row and column) for later usage:
		for(p=0;p<m[ispec].nlev;p++){
			partner[ipart].ctot[p]=0.;
			for(t=0;t<m[ispec].nlev;t++) partner[ipart].ctot[p]+=gsl_matrix_get(partner[ipart].colli,p,t);
		}
	}

	// Add the entries of the collision matrix to the total matrix:
	for(p=0;p<m[ispec].nlev;p++){
		// Diagonal values:
		for(ipart=0;ipart<m[ispec].npart;ipart++){	
	  		gsl_matrix_set(matrix,p,p,gsl_matrix_get(matrix,p,p)+g[id].dens[ipart]*partner[ipart].ctot[p]);
		}
		// Off-diagonal values:
		for(t=0;t<m[ispec].nlev;t++){
			if(p!=t){ 
				for(ipart=0;ipart<m[ispec].npart;ipart++){	
				  gsl_matrix_set(matrix,p,t,gsl_matrix_get(matrix,p,t)-g[id].dens[ipart]*gsl_matrix_get(partner[ipart].colli,t,p));
			  	}
			}	
		}
		// Set entries of last row of final matrix to unity:
		gsl_matrix_set(matrix, m[ispec].nlev, p, 1.);
		// Set entries of last column of final matrix to zero:
		gsl_matrix_set(matrix, p, m[ispec].nlev, 0.);		
	}

#ifdef DEBUG_MATRICES2FILE
	// Open the file:
	FILE *MatricesOut;
	MatricesOut = fopen("debug_matrices.out", "w");

	// Print collision matrices:
	for(ipart = 0; ipart < m[ispec].npart; ipart++){
		fprintf(MatricesOut, "partner[%d].colli:\n", ipart);
		for(l = 0; l < m[ispec].nlev+1; l++){
			fprintf(MatricesOut, "%9d", l);
		}
		fprintf(MatricesOut, "\n");
		for(k = 0; k < m[ispec].nlev+1; k++){
			fprintf(MatricesOut, "%d ", k);
			for(l = 0; l < m[ispec].nlev+1; l++){
				fprintf(MatricesOut, "%3.2e ", gsl_matrix_get(partner[ipart].colli,k,l));
			}
			fprintf(MatricesOut, "\n");
		}
		fprintf(MatricesOut, "\n");
	}

	// Print total matrix:
	fprintf(MatricesOut, "matrix:\n");
	for(l = 0; l < m[ispec].nlev+1; l++){
		fprintf(MatricesOut, "%10d", l);
	}
	fprintf(MatricesOut, "\n");
	for(k = 0; k < m[ispec].nlev+1; k++){
		fprintf(MatricesOut, "%d ", k);
		for(l = 0; l < m[ispec].nlev+1; l++){
			fprintf(MatricesOut, "%+3.2e ", gsl_matrix_get(matrix,k,l));
		}
		fprintf(MatricesOut, "\n");
	}
	fprintf(MatricesOut, "\n");

	// Close the file:
	fclose(MatricesOut);
	bail_out("STOP HERE - debug matrices\n");exit(1);
#endif

	// Free memory:
	for(ipart=0;ipart<m[ispec].npart;ipart++){
		gsl_matrix_free(partner[ipart].colli);
		free(partner[ipart].ctot);
	}
	free(partner);
}
